@@ -1,3 +1,16 @@
+/* 
+ * Copyright 2003,2004,2005 Kevin Smathers, All Rights Reserved
+ *
+ * This file may be used or modified without the need for a license.
+ *
+ * Redistribution of this file in either its original form, or in an
+ * updated form may be done under the terms of the GNU GENERAL
+ * PUBLIC LICENSE.  If this license is unacceptable to you then you
+ * may not redistribute this work.
+ * 
+ * See the file COPYING for details.
+ */
+
 /* btcheck.c */
 
 #include "config.h"
@@ -38,7 +51,7 @@ int main(int argc, char **argv)
 {
 	int result=0;
 	int dlid;
-	char *filename, *dirname;
+	char *filename;
 	btStream *bts;
 	struct btContext *ctx = &context;
 
@@ -46,13 +59,7 @@ int main(int argc, char **argv)
 	if(argc!=2) goto usage;
 	if(argv[1][0]=='-') goto usage;
 
-	dirname=argv[1];
-	filename=dirname+strlen(dirname);
-	while(filename>dirname&&filename[-1]!='/') --filename;
-	if(filename!=dirname) {
-		filename[-1]='\0';
-		chdir(dirname);
-	}
+	filename=argv[1];
 
 	if(!(bts=bts_create_filestream(filename, BTS_INPUT))) {
 		perror(filename);
@@ -73,6 +80,7 @@ int main(int argc, char **argv)
 
 usage:
 	fprintf(stderr, "Usage: %s torrentfile\n", argv[0]);
+	fprintf(stderr, "Version: %.2f\n", VERSION);
 	return 1;
 }
 
