@@ -1,8 +1,19 @@
+/* 
+ * Copyright 2003,2004,2005 Kevin Smathers, All Rights Reserved
+ *
+ * This file may be used or modified without the need for a license.
+ *
+ * Redistribution of this file in either its original form, or in an
+ * updated form may be done under the terms of the GNU LIBRARY GENERAL
+ * PUBLIC LICENSE.  If this license is unacceptable to you then you
+ * may not redistribute this work.
+ * 
+ * See the file COPYING.LGPL for details.
+ */
+
 # include "config.h"
 
-#if WIN32
-//#   include <winsock2.h>
-#else
+#if !WIN32
 #   include <sys/types.h>
 #   include <netinet/in.h>
 #   include <arpa/inet.h>
@@ -38,11 +49,17 @@
 #   define close(s) closesocket(s)
 #   define int32_t signed int
 #endif
-#define REQMAX 10
+#define REQMAX 10	    /* Maximum requests to send.  Must be less than queuesize or local queue will overflow */
 #define DOWNLOADS 4
 #define REQUEST_SIZE 16384  /* Default request size */
 char g_filebuffer[MAXREQUEST];	/* This should be moved to context or allocated in process_queue */
 
+#ifdef TRACE
+#   define bttrace(msg) printf("%s+%d:%s\n", __FILE__, __LINE__, msg)
+#else
+#   define bttrace(msg)
+#endif
+
 btPeerset* btPeerset_create( btPeerset *pset) {
     if (!pset) {
 	pset = (btPeerset *)btmalloc(sizeof(btPeerset));
@@ -60,13 +77,16 @@ int
 send_choke( btPeer *peer, int choke) ;
 
 int peer_connect_complete( btContext* ctx, struct btPeer *p) {
-    char error;
+    int error;
     int	errlen;
+    bttrace("peer_connect_complete");
+
 
     errlen = sizeof(error);
     if (getsockopt( p->ios.fd, SOL_SOCKET, SO_ERROR, &error, &errlen)) {
 	p->ios.error = errno;
 	bts_perror(p->ios.error, "getsockopt");
+       bttrace("peer_connect_complete error");
 	return p->ios.error;
     }
     if (error != 0) {
@@ -74,10 +94,12 @@ int peer_connect_complete( btContext* ctx, struct btPeer *p) {
 	p->ios.error = error;
 	bts_perror(p->ios.error, "connect_complete");
 	p->state = PEER_ERROR;
+       bttrace("peer_connect_complete error");
 	return p->ios.error;
     }
     p->state = PEER_OUTGOING;
     printf("%d: completed connection %s\n", p->ios.fd, inet_ntoa(p->ip));
+   bttrace("peer_connect_complete exit");
     return 0;    
 }
 
@@ -87,13 +109,22 @@ int peer_connect_request( btContext* ctx, struct btPeer *p) {
     long flags;
     int err;
 
-    if (p == NULL) return -1;
-    if (!memcmp(ctx->myid, p->id, IDSIZE)) return -1;
+    bttrace("peer_connect_request");
+
+    if (p == NULL) {
+	bttrace("peer_connect_request error");
+    	return -1;
+    }
+    if (!memcmp(ctx->myid, p->id, IDSIZE)) {
+	bttrace("peer_connect_request error");
+    	return -1;
+    }
 
     /* create the socket */
     sock = socket( PF_INET, SOCK_STREAM, 0);
     if (sock == -1) {
 	bts_perror(errno, "socket");
+	bttrace("peer_connect_request error");
 	return -1;
     }
 
@@ -102,18 +133,24 @@ int peer_connect_request( btContext* ctx, struct btPeer *p) {
     flags = ioctlsocket( sock, FIONBIO, (unsigned long *) 1);
     if (flags != 0) {
         bts_perror(errno, "ioctlsocket");
+	close(sock);
+	bttrace("peer_connect_request error");
         return -1;
     }
 #else
     flags = fcntl( sock, F_GETFL);
     if (flags < 0) {
         bts_perror(errno, "fcntl F_GETFL");
+	close(sock);
+	bttrace("peer_connect_request error");
         return -1;
     }
     
     flags |= O_NONBLOCK;
     if ( fcntl( sock, F_SETFL, flags)) {
         bts_perror(errno, "fcntl F_SETFL");
+	close(sock);
+	bttrace("peer_connect_request error");
         return -1;
     }
 #endif
@@ -131,6 +168,7 @@ int peer_connect_request( btContext* ctx, struct btPeer *p) {
 	if (errno != EINPROGRESS) {
 	    bts_perror(errno, "connect");
 	    close(sock);
+	    bttrace("peer_connect_request error");
 	    return -1;
 	}
     }
@@ -140,7 +178,12 @@ int peer_connect_request( btContext* ctx, struct btPeer *p) {
     /* index the peer by socket */
     ctx->sockpeer[sock]=p;
 
-    if (err) return 1;
+    if (err) {
+	/* EINPROGRESS */
+	bttrace("peer_connect_request error (EINPROGRESS)");
+	return 1;
+    }
+    bttrace("peer_connect_request exit");
     return 0;
 }
 
@@ -150,11 +193,13 @@ btPeer *peer_add( btContext *ctx, unsigned download, char *id, struct in_addr *i
     int idx, err;
     struct btPeer *p;
 
+    bttrace("peer_add");
     DIE_UNLESS(download<ctx->downloadcount);
 
     /* initialize status info */
     if (ctx_addstatus( ctx, TMPLOC)) {
 	/* over the connection limit */
+	bttrace("peer_add error");
 	return NULL;
     }
 
@@ -174,6 +219,14 @@ btPeer *peer_add( btContext *ctx, unsigned download, char *id, struct in_addr *i
     idx = pset->len++;
     err = peer_connect_request( ctx, p);
 /*    printf("peer_add(..., %s, %d) = %d\n", ip, port, idx); */
+    if (err < 0) {
+	/* error connecting, clean up and return */
+	ctx_delstatus( ctx, TMPLOC);
+/*	bts_perror(errno, "peer_connect_request failed"); */
+        btfree(p);
+	bttrace("peer_add error");
+        return NULL;
+    }
 
     /* index the peer by peerid */
     pset->peer = btrealloc( pset->peer, sizeof(struct btPeer*) * pset->len);
@@ -206,6 +259,7 @@ btPeer *peer_add( btContext *ctx, unsigned download, char *id, struct in_addr *i
 	}
     }
 
+    bttrace("peer_add exit");
     return p;
 }
 
@@ -217,18 +271,30 @@ int peer_send_handshake( btContext *ctx, btPeer *peer) {
     char shake[0x14] = "\x13" "BitTorrent protocol";
     char flags[8] = { 0 };
     int err;
-#if 0
-    printf("peer_send_handshake\n");
-#endif
+
+    bttrace("peer_send_handshake");
     assert(peer->download >= 0);
     err = kStream_fwrite( &peer->ios, shake, 0x14);
-    if (err < 0) return -1;
+    if (err < 0) {
+	bttrace("peer_send_handshake return -1");
+    	return -1;
+    }
     err = kStream_fwrite( &peer->ios, flags, 8);
-    if (err < 0) return -2;
+    if (err < 0) {
+	bttrace("peer_send_handshake return -2");
+    	return -2;
+    }
     err = kStream_fwrite( &peer->ios, ctx->downloads[peer->download]->infohash, SHA_DIGEST_LENGTH);
-    if (err < 0) return -3;
+    if (err < 0) {
+	bttrace("peer_send_handshake return -3");
+    	return -3;
+    }
     err = kStream_fwrite( &peer->ios, ctx->myid, IDSIZE);
-    if (err < 0) return -4;
+    if (err < 0) {
+	bttrace("peer_send_handshake return -4");
+    	return -4;
+    }
+    bttrace("peer_send_handshake exit");
     return 0;
 }
 
@@ -253,8 +319,11 @@ static int check_handshake( btContext *ctx, btPeer *peer, char *shake, int len)
     char *flags = shake + PROTO_LENGTH;
     char *infohash = flags + FLAGS_LENGTH;
     char *id = infohash + SHA_DIGEST_LENGTH;
+
+    bttrace( "check_handshake");
     if (shake[0] != '\x13') {
 	peer->ios.error = BTERR_PROTOCOL_ID;
+	bttrace( "check_handshake error");
 	return -1;
     }
     if (len >= PROTO_LENGTH 
@@ -262,6 +331,7 @@ static int check_handshake( btContext *ctx, btPeer *peer, char *shake, int len)
     {
 	/* bad protocol */
 	peer->ios.error = BTERR_PROTOCOL_ID;
+	bttrace( "check_handshake error");
 	return -1;
     }
     if (len >= PROTO_LENGTH + FLAGS_LENGTH 
@@ -299,6 +369,7 @@ static int check_handshake( btContext *ctx, btPeer *peer, char *shake, int len)
 		    if (peer_send_handshake( ctx, peer)<0 || peer_send_bitfield( ctx, peer)<0)
 		    {
 		        peer->state=PEER_ERROR;
+			bttrace( "check_handshake error");
 			return -1;
 		    }
 		    break;
@@ -308,19 +379,22 @@ static int check_handshake( btContext *ctx, btPeer *peer, char *shake, int len)
 	else
 	{
 	    i=peer->download;
-	    if (memcmp( infohash, ctx->downloads[i]->infohash, SHA_DIGEST_LENGTH)!=0)
+	    if (memcmp( infohash, ctx->downloads[i]->infohash, SHA_DIGEST_LENGTH)!=0) {
 	        i=INT_MAX;
+	    }
 	}
 	if(i>=ctx->downloadcount)
 	{
 	    /* bad infohash from the peer */
 	    peer->ios.error = BTERR_HASH_MISMATCH;
+	    bttrace( "check_handshake error");
 	    return -1;
 	}
     }
     if (len >= PROTO_LENGTH + FLAGS_LENGTH + SHA_DIGEST_LENGTH + IDSIZE) {
 	memcpy( peer->id, id, IDSIZE);
     }
+    bttrace( "check_handshake exit");
     return 0;
 }
 
@@ -343,6 +417,8 @@ static int check_handshake( btContext *ctx, btPeer *peer, char *shake, int len)
 int recv_handshake( btContext *ctx, btPeer *peer) {
     char shake[PROTO_LENGTH + FLAGS_LENGTH + SHA_DIGEST_LENGTH + IDSIZE];
     int err;
+
+    bttrace( "recv_handshake");
     err = kStream_fread( &peer->ios, shake, sizeof(shake));
     if (err < 0) {
 	if (peer->ios.error == EAGAIN) {
@@ -353,21 +429,26 @@ int recv_handshake( btContext *ctx, btPeer *peer) {
 		DIE_UNLESS(err == len);
 		if (check_handshake( ctx, peer, shake, len)) {
 		    peer->state = PEER_ERROR;
+		    bttrace( "recv_handshake error");
 		    return -1;
 		}
 	    }
 	    peer->ios.error = EAGAIN;
 	}
+	bttrace( "recv_handshake error");
 	return -1;
     }
     DIE_UNLESS(err == sizeof(shake));
     if (check_handshake( ctx, peer, shake, sizeof(shake))) {
+	bttrace( "recv_handshake error");
 	return -1;
     }
     peer->state = PEER_GOOD;
+    ctx->downloads[peer->download]->peerset.incomplete++;
 #if 0
     printf("%d: got handshake\n", peer->ios.fd);
 #endif
+    bttrace( "recv_handshake exit");
     return 0;
 }
 
@@ -378,18 +459,25 @@ static int
 queue_request( btRequestQueue *q, int piece, int offset, int len) {
     int oldtail = q->tail;
     int newtail = (oldtail+1) % QUEUESIZE;
+
+    bttrace("queue_request");
+
     if (len > MAXREQUEST) return -1;
     if (newtail == q->head) return -2;
     q->req[oldtail].block = piece;
     q->req[oldtail].offset = offset;
     q->req[oldtail].length = len;
     q->tail = newtail;
+
+    bttrace("queue_request exit");
     return 0;
 }
 
 static int 
 clear_request_queue( btRequestQueue *q) {
+    bttrace("clear_request_queue");
     q->head = q->tail;
+    bttrace("clear_request_queue exit");
     return 0;
 }
 
@@ -399,6 +487,8 @@ remove_queued_request( btRequestQueue *q, int block, int offset, int len) {
     btRequest *r;
     int i = q->head;
     int offby=0;
+
+    bttrace("remove_queued_request");
     while (i != q->tail) {
 	r = &q->req[i];
 
@@ -413,6 +503,8 @@ remove_queued_request( btRequestQueue *q, int block, int offset, int len) {
 	i = (i+1) % QUEUESIZE;
     }
     q->tail = (q->tail + QUEUESIZE - offby) % QUEUESIZE;
+
+    bttrace("remove_queued_request exit");
     return offby;
 }
 
@@ -420,45 +512,75 @@ static btRequest *dequeue_request( btRequestQueue *q) {
     btRequest *req;
     int oldhead = q->head;
     int newhead = (oldhead+1) % QUEUESIZE;
-    if (oldhead == q->tail) return NULL;
+
+    bttrace("dequeue_request");
+
+    if (oldhead == q->tail) {
+	bttrace("dequeue_request error");
+    	return NULL;
+    }
     req = &q->req[oldhead];
     q->head = newhead;
+    bttrace("dequeue_request exit");
     return req;
 }
 
 static int
 queue_len( btRequestQueue *q ) {
     int ct=q->tail - q->head;
+    bttrace("queue_len");
+
     if (ct < 0) ct += QUEUESIZE;
+
+    bttrace("queue_len exit");
     return ct;
 }
 
 static void start_rate_timer( btPeerStatus *ps, time_t now) {
+    bttrace("start_rate_timer");
     if (ps->send_time == 0) {
 	ps->send_time = now;
     }
+    bttrace("start_rate_timer exit");
 }
 
 static void stop_rate_timer( btPeerStatus *ps, time_t now) {
+    bttrace("stop_rate_timer");
     if (ps->send_time != 0) {
 	ps->total_time += now - ps->send_time;
 	ps->send_time = 0;
     }
+    bttrace("stop_rate_timer exit");
 }
 
 static int rate_timer( btPeerStatus *ps, time_t now) {
     int total;
+
+    bttrace("rate_timer");
+
     total = ps->total_time;
     if (ps->send_time != 0) {
 	total += now - ps->send_time;
     }
     if (total < 1) total = 1;
+
+    bttrace("rate_timer exit");
     return total;
 }
 
+static void shift_byte( void *ptr, int ival) {
+    void **p = (void **)ptr;
+    char *p1 = (char *)*p;
+
+    bttrace("shift_byte");
+    *p1++ = ival;
+    *p = p1;
+    bttrace("shift_byte exit");
+}
+
 #define SHIFT_INT32(ptr,nbo,ival) \
     (nbo=htonl(ival), memcpy(ptr,&nbo,sizeof(int32_t)), ptr+=sizeof(int32_t))
-#define SHIFT_BYTE(ptr,ival) ((*((unsigned char *)(ptr))++) = ival)
+#define SHIFT_BYTE(ptr,ival) (shift_byte(&ptr,ival))
 
 #define UNSHIFT_INT32(ptr,nbo,ival) \
     (memcpy(&nbo,ptr,sizeof(int32_t)), ival=ntohl(nbo), ptr+=sizeof(int32_t))
@@ -474,24 +596,30 @@ recv_peermsg( btContext *ctx, btPeer *peer) {
     btDownload *dl = ctx->downloads[peer->download];
     int32_t nbo_len;
     int len;
-    char msg[80];
+    char msg[84];
     char *nmsg, *param;
     int res = 0;
     int err;
     int32_t nbo;
 
+    bttrace("recv_peermsg");
+
     DIE_UNLESS(peer->download<ctx->downloadcount);
 
     err = kStream_fpeek( &peer->ios, (char *)&nbo_len, sizeof(nbo_len));
-    if (err < 0) return -1;
+    if (err < 0) { 
+	bttrace("recv_peermsg error");
+    	return -1;
+    }
     assert(err == sizeof(nbo_len));
     len = ntohl(nbo_len);
     if (len <= 80) {
 	nmsg = msg;
-    } else if (len <= MAXMESSAGE) {
+    } else if (len > 0 && len <= MAXMESSAGE) {
 	nmsg = btmalloc(len+sizeof(int32_t));
     } else {	/* Too big a packet, kill the peer. */
         peer->ios.error = BTERR_LARGEPACKET;
+	bttrace("recv_peermsg error");
 	return -1;
     }
 #if 0
@@ -501,6 +629,7 @@ recv_peermsg( btContext *ctx, btPeer *peer) {
 	/* keep alive */
 	err = kStream_fread( &peer->ios, (char *)&nbo_len, sizeof(nbo_len));
 	DIE_UNLESS(err == sizeof(nbo_len));
+	bttrace("recv_peermsg error");
 	return 0;
     }
     DIE_UNLESS(len <= MAXMESSAGE && len >= 0);
@@ -557,10 +686,16 @@ recv_peermsg( btContext *ctx, btPeer *peer) {
 	case BT_MSG_INTERESTED:
 	    peer->remote.interested = 1;
 	    ctx->downloads[peer->download]->peerset.interestedpeers++;
+	    if (!peer->local.choked) {
+		start_rate_timer( &peer->local, time(NULL));
+	    }
 	    break;
 	case BT_MSG_NOTINTERESTED:
 	    peer->remote.interested = 0;
 	    ctx->downloads[peer->download]->peerset.interestedpeers--;
+	    if (!peer->local.choked) {
+		stop_rate_timer( &peer->local, time(NULL));
+	    }
 	    break;
 	case BT_MSG_HAVE:
 	    {
@@ -569,6 +704,12 @@ recv_peermsg( btContext *ctx, btPeer *peer) {
 
 		UNSHIFT_INT32(param,nbo,block);
 		bs_set( &peer->blocks, block);
+		if (bs_isFull( &peer->blocks)) {
+		    if (peer->remote.complete == 0) { 
+			peer->remote.complete = 1;
+			dl->peerset.incomplete--;
+		    }
+		}
 		interest = update_interested( ctx, peer);
 		if (interest && 
 			!peer->remote.snubbed && 
@@ -581,6 +722,12 @@ recv_peermsg( btContext *ctx, btPeer *peer) {
 	    break;
 	case BT_MSG_BITFIELD:
 	    kBitSet_readBytes( &peer->blocks, nmsg+5, len-1);
+	    if (peer->blocks.nbits == bs_countBits( &peer->blocks)) {
+	        if (peer->remote.complete == 0) { 
+		    peer->remote.complete = 1;
+                    dl->peerset.incomplete--;
+		}
+	    }
 	    update_interested( ctx, peer);
 	    break;
 	case BT_MSG_REQUEST:
@@ -599,7 +746,11 @@ recv_peermsg( btContext *ctx, btPeer *peer) {
 
 		if (!bs_isSet(&peer->blocks, piece) && 
 		      bs_isSet(&ctx->downloads[peer->download]->fileset.completed, piece)) {
-		    queue_request( &peer->queue, piece, offs, len);
+		    res = queue_request( &peer->queue, piece, offs, len);
+		    if (res) {
+			peer->ios.error = BTERR_QUEUE_OVERFLOW;
+		        goto cleanup;
+		    }
 		} else {
 		    time_t now;
 		    time(&now);
@@ -627,7 +778,7 @@ recv_peermsg( btContext *ctx, btPeer *peer) {
 			offs,
 			len-9);
 #endif
-		/* FIXME: check requests with ones we've sent */
+		/* check requests with ones we've sent */
 		time(&peer->lastreceived);
 		if (!pp || !remove_queued_request( &peer->inqueue, piece, offs, len-9)) {
 		    printf("%d: Unneeded data: piece %d %d+%d\n", peer->ios.fd, piece, offs, len-9);
@@ -657,8 +808,7 @@ recv_peermsg( btContext *ctx, btPeer *peer) {
 					    req->length
 					)) {
 					p->state = PEER_ERROR;
-					err = -1;
-					goto cleanup;
+					continue;
 				    }
 				}
 
@@ -667,8 +817,7 @@ recv_peermsg( btContext *ctx, btPeer *peer) {
 			    if (p->state == PEER_GOOD) {
 				if (send_have( p, piece)) {
 				    p->state = PEER_ERROR;
-				    err = -1;
-				    goto cleanup;
+				    continue;
 				}
 			    }
 			} /* for i in peerset.len */
@@ -717,6 +866,12 @@ cleanup:
 #if 0
 	    printf("B/looking for %d bytes %d available (addr %d)\n", tlen, kStream_iqlen( &peer->ios), kStream_in_addr(&peer->ios));
 #endif
+            if (tlen < 0 || tlen > MAXMESSAGE) {
+		/* out of sync with peer, or packet too large */
+		peer->ios.error = BTERR_LARGEPACKET;
+		bttrace("recv_peermsg error");
+		return -1;
+	    }
 	    DIE_UNLESS(tlen <= MAXMESSAGE && tlen >= 0);
 	    if (kStream_iqlen( &peer->ios) >= tlen) res = 1;
 	}
@@ -725,6 +880,7 @@ cleanup:
     if (len > 80) {
 	btfree(nmsg);
     }
+    bttrace("recv_peermsg exit");
     return res;
 }
 
@@ -738,43 +894,65 @@ cleanup:
 int
 send_message( btPeer *peer, char *buf, int len) {
     int nslen = htonl( len);
+
+    bttrace("send_message");
 #if 0
 	printf("%d: send message %d len=%d addr=%d\n", peer->ios.fd, buf?buf[0]:-1, len, kStream_out_addr(&peer->ios));
 #endif
     if (kStream_fwrite( &peer->ios, (void*)&nslen, sizeof(nslen)) < 0) {
+	bttrace("send_message error");
 	return -1;
     }
     if (len > 0) {
 	if (kStream_fwrite( &peer->ios, buf, len) < 0) {
+	    bttrace("send_message error");
 	    return -2;
 	}
     }
+    bttrace("send_message exit");
     return 0;
 }
 
 int
 send_keepalive( btPeer *peer) {
-    return send_message( peer, NULL, 0);
+    int res;
+    bttrace("send_keepalive entry");
+    res = send_message( peer, NULL, 0);
+    bttrace("send_keepalive exit");
+    return res;
 }
 
 int
 send_choke( btPeer *peer, int choke) {
     char type=choke?BT_MSG_CHOKE:BT_MSG_UNCHOKE;
+    int res;
+
+    bttrace("send_choke");
 
     if (peer->local.choked != choke) {
 	/* cancel all requests */
 	clear_request_queue( &peer->queue);
 	peer->local.choked = choke;
     }
-    return send_message( peer, &type, 1);
+
+    res=send_message( peer, &type, 1);
+
+    bttrace("send_choke exit");
+    return res;
 }
 
 int
 send_interested( btPeer *peer, int interest) {
     char type=interest?BT_MSG_INTERESTED:BT_MSG_NOTINTERESTED;
+    int res;
+
+    bttrace("send_interested");
 
     peer->local.interested = interest;
-    return send_message( peer, &type, 1);
+    res=send_message( peer, &type, 1);
+
+    bttrace("send_interested exit");
+    return res;
 }
 
 
@@ -783,9 +961,15 @@ send_have( btPeer *peer, int piece) {
     char buf[5];
     char *p=buf;
     int32_t nbo;
+    int res;
+
+    bttrace("send_have");
     SHIFT_BYTE( p, BT_MSG_HAVE);
     SHIFT_INT32( p, nbo, piece);
-    return send_message( peer, buf, p-buf);
+    res = send_message( peer, buf, p-buf);
+
+    bttrace("send_have exit");
+    return res;
 }
 
 int 
@@ -795,6 +979,8 @@ send_bitfield( btPeer *peer, kBitSet *set) {
     int res;
     int nbytes = (set->nbits + 7)/8;
 
+    bttrace("send_bitfield");
+
     nslen = htonl( 1 + nbytes);
     res = kStream_fwrite( &peer->ios, (void *)&nslen, sizeof(nslen));
     if (res < 0) return -1;
@@ -802,6 +988,8 @@ send_bitfield( btPeer *peer, kBitSet *set) {
     if (res < 0) return -2;
     res = kStream_fwrite( &peer->ios, set->bits, nbytes);
     if (res < 0) return -3;
+
+    bttrace("send_bitfield exit");
     return res;
 }
 
@@ -810,6 +998,10 @@ send_request( btPeer *peer, int piece, int offs, int len) {
     char buf[13];
     char *p=buf;
     int32_t nbo;
+    int res;
+
+    bttrace("send_request");
+
     SHIFT_BYTE( p, BT_MSG_REQUEST);
     SHIFT_INT32( p, nbo, piece);
     SHIFT_INT32( p, nbo, offs);
@@ -817,7 +1009,10 @@ send_request( btPeer *peer, int piece, int offs, int len) {
 #if 0
     printf("%d: send_request( ..., %d, %d, %d)\n", peer->ios.fd, piece, offs, len);
 #endif
-    return send_message( peer, buf, p-buf);
+    res = send_message( peer, buf, p-buf);
+
+    bttrace("send_request exit");
+    return res;
 }
 
 int 
@@ -828,6 +1023,9 @@ send_piece( btPeer *peer, int piece, int offs, char* cbuf, int len) {
     char *p=buf;
     int res;
     int32_t nbo;
+
+    bttrace("send_piece");
+
     SHIFT_BYTE( p, BT_MSG_PIECE);
     SHIFT_INT32( p, nbo, piece);
     SHIFT_INT32( p, nbo, offs);
@@ -842,6 +1040,8 @@ send_piece( btPeer *peer, int piece, int offs, char* cbuf, int len) {
     if (res < 0) return -2;
     res = kStream_fwrite( &peer->ios, cbuf, len);
     if (res < 0) return -3;
+
+    bttrace("send_piece exit");
     return res;
 }
 
@@ -849,18 +1049,31 @@ int
 process_queue( btFileSet *fs, btPeer *peer) {
     btRequest *req;
     int err;
+
+    bttrace("process_queue");
+
     req = dequeue_request( &peer->queue);
-    if (!req) return 0;
+    if (!req) {
+	bttrace("process_queue exit");
+	return 0;
+    }
 
     /* send the requested block */
     err = seg_readbuf( fs, req->block, req->offset, g_filebuffer, req->length);
-    if (err < 0) return err;
+    if (err < 0) {
+	bttrace("process_queue error");
+    	return err;
+    }
     
     err = send_piece( peer, req->block, req->offset, g_filebuffer, req->length);
-    if (err < 0) return err;
+    if (err < 0) {
+	bttrace("process_queue error");
+	return err;
+    }
 
     fs->ul += req->length;
 
+    bttrace("process_queue exit");
     return 1;
 }
 
@@ -869,6 +1082,9 @@ send_cancel( btPeer *peer, int piece, int offs, int len) {
     char buf[13];
     char *p=buf;
     int32_t nbo;
+    int res;
+
+    bttrace("send_cancel");
     SHIFT_BYTE( p, 8);
     SHIFT_INT32( p, nbo, piece);
     SHIFT_INT32( p, nbo, offs);
@@ -876,7 +1092,10 @@ send_cancel( btPeer *peer, int piece, int offs, int len) {
 #if 0
     printf("%d: send_cancel (%d, %d + %d)\n", peer->ios.fd, piece, offs, len);
 #endif
-    return send_message( peer, buf, p-buf);
+    res = send_message( peer, buf, p-buf);
+
+    bttrace("send_cancel exit");
+    return res;
 }
 
 #if 0
@@ -898,10 +1117,13 @@ int peer_answer( btContext *ctx, int sock) {
     struct hostent *hent;
 #endif
 
+    bttrace("peer_answer");
+
     /* allocate status */
     if (ctx_addstatus( ctx, sock)) {
 	/* over the connection limit */
 	close( sock);
+	bttrace("peer_answer error");
         return -1;
     }
 
@@ -942,23 +1164,28 @@ int peer_answer( btContext *ctx, int sock) {
     flags = ioctlsocket( sock, FIONBIO, (unsigned long *) 1);
     if (flags != 0) {
         bts_perror(errno, "ioctlsocket");
+	bttrace("peer_answer error");
         return -1;
     }
 #else
     flags = fcntl( sock, F_GETFL);
     if (flags < 0) {
         bts_perror(errno, "fcntl F_GETFL");
+	bttrace("peer_answer error");
         return -1;
     }
     
     flags |= O_NONBLOCK;
     if ( fcntl( sock, F_SETFL, flags)) {
         bts_perror(errno, "fcntl F_SETFL");
+	bttrace("peer_answer error");
         return -1;
     }
 #endif
 
     ctx_setevents( ctx, sock, POLLIN);
+
+    bttrace("peer_answer exit");
     return 0;
 }
 
@@ -975,13 +1202,20 @@ int peer_answer( btContext *ctx, int sock) {
  */
 
 int peer_process_queue( btFileSet *fs, btPeer *p) {
+    int res;
+
+    bttrace("peer_process_queue");
     if ( kStream_oqlen(&p->ios) <= 8096 ) {
 	/* output buffer has drained, grab next request from queue */
-	return process_queue( fs, p);
+	res = process_queue( fs, p);
+
+	bttrace("peer_process_queue exit");
+	return res;
     } else {
 #if 0
 	printf("%d: kStream_oqlen(&p->ios) = %d\n", p->ios.fd, kStream_oqlen(&p->ios));
 #endif
+	bttrace("peer_process_queue exit");
 	return 0;
     }
 }
@@ -989,6 +1223,7 @@ int peer_process_queue( btFileSet *fs, btPeer *p) {
 void peer_shutdown( btContext *ctx, btPeer *peer, char *error) {
     int fd = peer->ios.fd;
 
+    bttrace("peer_shutdown");
 #if 0
     printf("%d: peer_shutdown()\n", fd);
 #endif
@@ -1012,6 +1247,7 @@ void peer_shutdown( btContext *ctx, btPeer *peer, char *error) {
 
     kStream_finit( &peer->ios);
     kBitSet_finit( &peer->blocks);
+    bttrace("peer_shutdown exit");
 }    
 
 /*
@@ -1024,6 +1260,8 @@ void peer_shutdown( btContext *ctx, btPeer *peer, char *error) {
 int 
 peer_recv_message( btContext *ctx, btPeer *p) {
     int res=0;
+
+    bttrace("peer_recv_message");
 #if 0
     printf("%d: before message state %d\n", p->ios.fd, p->state);
 #endif
@@ -1036,6 +1274,7 @@ peer_recv_message( btContext *ctx, btPeer *p) {
 		stat, ctx->status[ stat].events, ctx->status[ stat].revents,
 		ctx->status[ stat].fd);
 #endif
+	bttrace("peer_recv_message exit");
 	return 0;
     } else if (p->state == PEER_OUTGOING || p->state == PEER_INCOMING) {
 	res = recv_handshake( ctx, p);
@@ -1052,12 +1291,15 @@ peer_recv_message( btContext *ctx, btPeer *p) {
 #if 0
     printf("%d: after message state %d\n", p->ios.fd, p->state);
 #endif
+    bttrace("peer_recv_message exit");
     return res;
 }
 
 int
 update_interested( btContext *ctx, btPeer *p) {
     int interest;
+
+    bttrace("update_interested");
     btDownload *dl=ctx->downloads[p->download];
     DIE_UNLESS (p->download<ctx->downloadcount);
     interest = bs_hasInteresting( &dl->fileset.completed, &p->blocks, &dl->interested);
@@ -1068,26 +1310,48 @@ update_interested( btContext *ctx, btPeer *p) {
 	    stop_rate_timer( &p->remote, time(NULL));
 	}
     } 
+    bttrace("update_interested exit");
     return interest;
 }
 
+static int countpeers( btPeerset *peers, btPartialPiece *piece) {
+    int count = 0;
+    int i;
+
+    bttrace("countpeers");
+    for (i=0; i < peers->len; i++) {
+        btPeer *p = peers->peer[i];
+	if (p->currentPiece == piece) {
+	    count ++;
+	}
+    }
+    bttrace("countpeers exit");
+    return count;
+}
+
+
 btPartialPiece * 
 peer_assign_block( struct btContext *ctx, btPeer *p) {
     btPartialPiece *piece;
     btDownload *dl=ctx->downloads[p->download];
+
+    bttrace("peer_assign_block");
     DIE_UNLESS (p->download<ctx->downloadcount);
 
     piece=dl->fileset.partial;
-    if (dl->peerset.interestedpeers > 4) {
-        /* need blocks quickly; over-request to get a new block */
+    if (dl->peerset.interestedpeers > 4 ||
+        dl->peerset.incomplete <= 4 ||
+        dl->peerset.len - dl->peerset.incomplete > 10) 
+    {
+        /* assign a partial block if no one else is working on it */
 	while(piece && (bs_isSet(&dl->requested, piece->piecenumber) ||
 	      !bs_isSet( &p->blocks, piece->piecenumber))) {
 	    piece=piece->next;
 	}
 	if (piece) printf("%d: assigning partial block %d\n", p->ios.fd, piece->piecenumber);
     } else {
-	/* avoid overlapping requests since we have interested peers */
-	while(piece && !bs_isSet( &p->blocks, piece->piecenumber)) {
+	/* use overlapping requests to get more interested peers (but not more than three on a block) */
+	while(piece && !bs_isSet( &p->blocks, piece->piecenumber) && countpeers(&dl->peerset, piece)<4) {
 	    piece=piece->next;
 	}
 	if (piece) printf("%d: assigning overlapping block %d\n", p->ios.fd, piece->piecenumber);
@@ -1103,6 +1367,7 @@ peer_assign_block( struct btContext *ctx, btPeer *p) {
     }
     p->currentPiece = piece;
     bs_set( &dl->requested, piece->piecenumber);
+    bttrace("peer_assign_block exit");
     return piece;
 }
 
@@ -1113,11 +1378,13 @@ peer_assign_block( struct btContext *ctx, btPeer *p) {
 static int in_rate( btPeer *a, time_t now) {
     float atime, arate;
 	int newpeer;
+
+    bttrace("in_rate");
     atime = (float)rate_timer( &a->remote, now);
     arate = a->ios.read_count / atime;
 
     newpeer = rate_timer( &a->local, now);
-    if (newpeer < 30) {
+    if (newpeer < 30 && a->remote.interested) {
 	/* new peers start at 4k/s assumed rate for first 30s */
 	arate = (float)NEWPEER_LEVEL;
     } else if (atime < 30) {
@@ -1133,13 +1400,19 @@ static int in_rate( btPeer *a, time_t now) {
 	/* remote has us choked, or unchoked but we aren't interested */
 	arate *= (float)CHOKED_PENALTY;
     }
+    bttrace("in_rate exit");
     return (int)arate;
 }
 
 static int out_rate( btPeer *a, time_t now) {
     int atime, arate;
+
+    bttrace("out_rate");
+
     atime = rate_timer( &a->local, now);
     arate = a->ios.write_count / atime;
+
+    bttrace("out_rate exit");
     return arate;
 }
 
@@ -1154,6 +1427,7 @@ peer_send_request( btContext *ctx, btPeer *p) {
     int arate = in_rate(p, now);
     int qlen;
     btDownload *dl=ctx->downloads[p->download];
+    bttrace("peer_send_request");
 
     DIE_UNLESS(p->download<ctx->downloadcount);
 
@@ -1164,6 +1438,7 @@ peer_send_request( btContext *ctx, btPeer *p) {
     	piece = peer_assign_block( ctx, p);
         if (!piece) {
             update_interested( ctx, p);
+	    bttrace("peer_send_request exit");
             return 0;
         }
 #if 0
@@ -1189,15 +1464,22 @@ peer_send_request( btContext *ctx, btPeer *p) {
                 /* restart from unreceived */
                 int rst;
 
-		if (queue_len( &p->inqueue) > 0) return 0;
+		if (queue_len( &p->inqueue) > 0) {
+		    bttrace("peer_send_request exit");
+		    return 0;
+		}
                 rst = bs_firstClr( &piece->filled);
-                if (rst < 0) return 0;
+                if (rst < 0) {
+		    bttrace("peer_send_request exit");
+		    return 0;
+		}
                 piece->nextbyteReq = rst;
 	    }
 	    start = piece->nextbyteReq;
 	    len = blocklen - start;
 
 	    if (len == 0) {
+		bttrace("peer_send_request exit");
 	        return 0;
 	    }
 	}
@@ -1209,19 +1491,29 @@ peer_send_request( btContext *ctx, btPeer *p) {
 	res = send_request(p, piece->piecenumber, start, len);
 	queue_request( &p->inqueue, piece->piecenumber, start, len);
     }
+    bttrace("peer_send_request exit");
     return res;
 }
 
 int 
 peer_send_bitfield( btContext *ctx, btPeer *peer) {
+    int res;
+
+    bttrace("peer_send_bitfield");
 #if 0
     printf("peer_send_bitfield\n");
 #endif
-    return send_bitfield( peer, &ctx->downloads[peer->download]->fileset.completed);
+    res = send_bitfield( peer, &ctx->downloads[peer->download]->fileset.completed);
+    
+    bttrace("peer_send_bitfield exit");
+    return res;
 }
 
 static int compare_rate( btContext *ctx, btPeer *a, btPeer *b, time_t now) {
     int arate, brate;
+    int res = 0;
+
+    bttrace("compare_rate");
 
     if (ctx->downloads[a->download]->fileset.left == 0) {
 	arate = out_rate( a, now);
@@ -1230,32 +1522,55 @@ static int compare_rate( btContext *ctx, btPeer *a, btPeer *b, time_t now) {
 	arate = in_rate( a, now);
 	brate = in_rate( b, now);
     }
-    if (arate > brate) return -1;
-    if (arate < brate) return 1;
-    return 0;
+    if (arate > brate) res = -1;
+    if (arate < brate) res = 1;
+
+    bttrace("compare_rate exit");
+    return res;
 }
 
 static void prioritize( btContext *ctx, btPeer *p[DOWNLOADS], btPeer *new, time_t now) {
     int i,j;
+    
+    bttrace("prioritize");
     for (i=0; i<DOWNLOADS; i++) {
 	if (!new->local.snubbed && new->remote.interested) 
 	{
-	    if (p[i] == NULL || compare_rate( ctx, new, p[i], now) < 0) {
+	    if (p[i] == NULL || p[i]->remote.interested==0 || compare_rate( ctx, new, p[i], now) < 0) {
 		for (j=DOWNLOADS-1; j>i; j--) {
 		    p[j]=p[j-1];
 		}
 		p[i]=new;
+		bttrace("prioritize exit");
 		return;
 	    }
 	}
     }
+
+    for (i=0; i<DOWNLOADS; i++) {
+	if (p[i] == NULL || (p[i]->remote.interested==0 && compare_rate( ctx, new, p[i], now) < 0)) {
+	    for (j=DOWNLOADS-1; j>i; j--) {
+	    	p[j]=p[j-1];
+	    }
+	    p[i]=new;
+	    bttrace("prioritize exit");
+	    return;
+	}
+    }
+    bttrace("prioritize exit");
 }
 
 static int isfavorite( btPeer *p[DOWNLOADS], btPeer *check) {
     int i;
+
+    bttrace("isfavorite");
     for (i=0; i<DOWNLOADS; i++) {
-	if (p[i] == check) return 1;
+	if (p[i] == check) {
+	    bttrace("isfavorite exit");
+	    return 1;
+	}
     }
+    bttrace("isfavorite exit");
     return 0;
 }
 
@@ -1264,6 +1579,8 @@ void peer_favorites( btContext *ctx, btPeerset *pset) {
     time_t now;
     btPeer *p[DOWNLOADS] = { NULL };
 
+    bttrace("peer_favorites");
+
     time(&now);
 
     /* select the peers we will let download */
@@ -1283,7 +1600,9 @@ void peer_favorites( btContext *ctx, btPeerset *pset) {
 		if (send_choke( peer, 0)) {
 		    peer->state = PEER_ERROR;
 		}
-		start_rate_timer( &peer->local, now);
+		if (peer->remote.interested) {
+		    start_rate_timer( &peer->local, now);
+		}
 	    } 
 	} else {
 	    if (!peer->local.choked && peer->state == PEER_GOOD) {
@@ -1295,11 +1614,14 @@ void peer_favorites( btContext *ctx, btPeerset *pset) {
 	    }
 	}
     }
+    bttrace("peer_favorites exit");
 }
 
 void
 peer_status_dump( btPeerStatus *ps, int bytes) {
     int total, rate;
+
+    bttrace("peer_status_dump");
     total = rate_timer( ps, time(NULL));
     
     if (total == 0) total = 1;
@@ -1318,11 +1640,15 @@ peer_status_dump( btPeerStatus *ps, int bytes) {
     } else {
 	printf("(%3dbps)", rate);
     }
+    bttrace("peer_status_dump exit");
 }
 
 void
 peer_dump( btPeerset *pset) {
     int i;
+
+    bttrace("peer_dump");
+
     for (i=0; i<pset->len; i++) {
 	btPeer *p = pset->peer[i];
 	if (!p || p->state == PEER_ERROR) continue;
@@ -1350,6 +1676,7 @@ peer_dump( btPeerset *pset) {
 	peer_status_dump( &p->local, p->ios.write_count);
 	printf("_%d]\n", queue_len(&p->queue));
     }
+    bttrace("peer_dump exit");
 }
 
 void 
@@ -1359,6 +1686,9 @@ peer_summary( btPeerset *pset) {
     float rtime, ttime;
     float rbytes, tbytes;
     float rrate = 0, trate = 0;
+
+    bttrace("peer_summary");
+
     for (i=0; i<pset->len; i++) {
 	btPeer *p = pset->peer[i];
 	if (p == NULL) continue;
@@ -1369,7 +1699,7 @@ peer_summary( btPeerset *pset) {
 	    rrate += rbytes / rtime;
 	}
 
-	if (!p->local.choked) {
+	if (!p->local.choked && p->remote.interested) {
 	    ttime = (float)rate_timer( &p->local, time(NULL));
 	    tbytes = (float)p->ios.write_count;
 	    trate += tbytes / ttime;
@@ -1394,6 +1724,7 @@ peer_summary( btPeerset *pset) {
     }
     printf("\r");
     fflush(stdout);
+    bttrace("peer_summary exit");
 }
 
 /*
@@ -1402,17 +1733,22 @@ peer_summary( btPeerset *pset) {
 int
 peer_allcomplete( btPeerset *pset) {
   int i;
+
+  bttrace("peer_allcomplete");
   for (i=0; i<pset->len; i++) {
     btPeer *p = pset->peer[i];
     if (p->state == PEER_ERROR)
       continue;
     if (p->state != PEER_GOOD) {
       /* Something in progress */
+      bttrace("peer_allcomplete exit");
       return 0;
     }
     if (!bs_isFull (&p->blocks)) {
+      bttrace("peer_allcomplete exit");
       return 0;
     }
   }
+  bttrace("peer_allcomplete exit");
   return 1;
 }
